#ident	"@(#)slipd.c	1.2"
/*
 * Copyright (c) 1993 Addison-Wesley Publishing Company.
 * All rights reserved.  This source code is provided
 * for its instructional value, and is not guaranteed
 * for any particular purpose.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <net/if.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stropts.h>
#include <poll.h>
#include <wait.h>
#include <netdb.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <slipcmd.h>

typedef struct seriallink slink_t;

struct seriallink {
	slink_t	*next;		/* next entry in list */
	slink_t	*prev;		/* previous entry in list */
	dev_t	dev;		/* dev # of driver at end of stream */
	int	id;		/* link ID */
	char	hostname[MAXHOSTNAMELEN]; /* foreign host */
};

static slink_t *slist;
static int netfd;

static void svc_pipe(int);
static void svc_driver(int);
static int doconn(int, struct slcommand *);
static int dodiscon(struct slcommand *);
static int setifname(int);
static int ifup(int, struct conninfo *);
static int ifdown(int);
static int _discon(slink_t *);
static void timeout(int);

void
main()
{
	int fd, fl;
	int pfd[2];
	struct pollfd pary[2];
	struct sigaction sa;

	daemonize("slipd");
	fd = open(SLIPDEV, O_RDWR|O_NONBLOCK);
	if (fd < 0) {
		log("can't open %s: %s", SLIPDEV, strerror(errno));
		exit(1);
	}
	netfd = open("/dev/ip", O_RDWR);
	if (netfd < 0) {
		log("can't open /dev/ip: %s", strerror(errno));
		exit(1);
	}
	if (pipe(pfd) < 0) {
		log("can't create pipe: %s", strerror(errno));
		exit(1);
	}
	if ((fl = fcntl(pfd[1], F_GETFL, 0)) < 0) {
		log("can't get pipe file flags: %s", strerror(errno));
		exit(1);
	}
	if (fcntl(pfd[1], F_SETFL, fl|O_NONBLOCK) < 0) {
		log("can't set pipe file flags: %s", strerror(errno));
		exit(1);
	}
	if (ioctl(pfd[0], I_PUSH, "connld") < 0) {
		log("can't push CONNLD: %s", strerror(errno));
		exit(1);
	}
	close(creat(PIPEPATH, PIPEMODE));
	if (fattach(pfd[0], PIPEPATH) < 0) {
		log("can't attach pipe to %s: %s", PIPEPATH,
		  strerror(errno));
		exit(1);
	}
	sigemptyset(&sa.sa_mask);
	sa.sa_handler = timeout;
	sa.sa_flags = 0;
	sigaction(SIGALRM, &sa, NULL);
	pary[0].events = POLLIN;
	pary[0].fd = fd;
	pary[1].events = POLLIN;
	pary[1].fd = pfd[1];
	while (poll(pary, 2, INFTIM) >= 0) {
		if (pary[0].revents & (POLLERR|POLLHUP|POLLNVAL)) {
			log("poll driver revents %x", pary[0].revents);
			exit(1);
		}
		if (pary[1].revents & (POLLERR|POLLHUP|POLLNVAL)) {
			log("poll pipe revents %x", pary[1].revents);
			exit(1);
		}
		if (pary[0].revents & POLLIN)
			svc_driver(pary[0].fd);
		if (pary[1].revents & POLLIN)
			svc_pipe(pary[1].fd);
	}
	log("poll failed: %s", strerror(errno));
	exit(1);
}

static void
svc_pipe(int pipefd)
{
	struct strrecvfd piprecv;
	struct slcommand slcmd;
	int n, err;

	while (ioctl(pipefd, I_RECVFD, &piprecv) == 0) {
		alarm(30);
		n = read(piprecv.fd, &slcmd,
		  sizeof(struct slcommand));
		alarm(0);
		if (n != sizeof(struct slcommand)) {
			if (n < 0)
				log("can't read from received pipe: %s",
				  strerror(errno));
			else
				log("only read %d bytes from pipe", n);
		} else {
			err = 0;
			switch (slcmd.opcode) {
			case CONNECT:
				log("received connect request");
				if (doconn(piprecv.fd, &slcmd) < 0)
					err = errno;
				break;

			case DISCONNECT:
				log("received disconnect request");
				if (dodiscon(&slcmd) < 0)
					err = errno;
				break;

			default:
				err = EINVAL;
				log("unknown opcode: %d", slcmd.opcode);
			}
			write(piprecv.fd, &err, sizeof(int));
		}
		close(piprecv.fd);
	}
	if (errno == EAGAIN)
		return;
	log("error receiving fd on mounted pipe: %s",
	  strerror(errno));
}

static int
doconn(int pipefd, struct slcommand *cmdp)
{
	struct strrecvfd slrecv;
	struct conninfo *cip;
	struct stat sbuf;
	slink_t *slp;
	int id;

	cip = getcinfo(cmdp->arg);
	if (cip == NULL) {
		log("can't find info on host %s", cmdp->arg);
		return(-1);
	}
	alarm(30);
	if (ioctl(pipefd, I_RECVFD, &slrecv) < 0) {
		log("can't receive slfd: %s", strerror(errno));
		return(-1);
	}
	alarm(0);
	slp = NULL;
	if (fstat(slrecv.fd, &sbuf) < 0) {
		log("can't stat slfd: %s", strerror(errno));
		goto err;
	}
	slp = (slink_t *)malloc(sizeof(slink_t));
	if (slp == NULL) {
		log("can't allocate slink_t structure: %s",
		  strerror(errno));
		goto err;
	}
	if (ioctl(slrecv.fd, I_PUSH, "slip") < 0) {
		log("can't push slip module: %s", strerror(errno));
		goto err;
	}
	id = ioctl(netfd, I_PLINK, slrecv.fd);
	if (id < 0) {
		log("link failed: %s", strerror(errno));
		goto err;
	}
	if (setifname(id) < 0) {
		log("can't set interface name: %s", strerror(errno));
		ioctl(netfd, I_PUNLINK, id);
		goto err;
	}
	if (ifup(id, cip) < 0) {
		log("can't ifconfig");
		ioctl(netfd, I_PUNLINK, id);
		goto err;
	}
	slp->dev = sbuf.st_rdev;
	slp->id = id;
	slp->next = slist;
	slp->prev = NULL;
	strcpy(slp->hostname, cip->fhost);
	if (slist != NULL)
		slist->prev = slp;
	slist = slp;
	log("link dev 0x%x ID %d created", slp->dev, id);
	close(slrecv.fd);
	return(0);
err:
	close(slrecv.fd);
	if (slp != NULL)
		free(slp);
	return(-1);
}

static int
setifname(int id)
{
	struct strioctl strioc;
	struct ifreq req;

	sprintf(req.ifr_name, "sl%d", id);
	req.ifr_metric = id;
	strioc.ic_cmd = SIOCSIFNAME;
	strioc.ic_timout = INFTIM;
	strioc.ic_len = sizeof(struct ifreq);
	strioc.ic_dp = (caddr_t)&req;
	return(ioctl(netfd, I_STR, &strioc));
}

static int
ifup(int id, struct conninfo *cip)
{
	pid_t pid;
	int status;
	char ifname[IFNAMSIZ];

	sprintf(ifname, "sl%d", id);
	if ((pid = fork()) < 0) {
		log("fork failed: %s", strerror(errno));
		return(-1);
	}
	if (pid == 0) {		/* child */
		execl(IFCONFIGPATH, IFCONFIG, ifname, "inet",
		  cip->laddr, cip->faddr, "up", NULL);
		log("unexpected return from exec");
		exit(1);
	}
	/* parent */
	while (waitpid(pid, &status, 0) != pid)
		;
	if (WIFEXITED(status) && (WEXITSTATUS(status) == 0)) {
		return(0);
	} else {
		errno = EPROTO; /* as good as any */
		return(-1);
	}
}

static int
dodiscon(struct slcommand *cmdp)
{
	slink_t *slp;

	for (slp = slist; slp != NULL; slp = slp->next)
		if (strcmp(slp->hostname, cmdp->arg) == 0)
			break;
	if (slp == NULL) {
		log("no slink_t structure for host %s", cmdp->arg);
		errno = EINVAL;
		return(-1);
	}
	return(_discon(slp));
}

static int
_discon(slink_t *slp)
{
	if (ifdown(slp->id) < 0) {
		log("can't mark interface down");
		return(-1);
	}
	if (ioctl(netfd, I_PUNLINK, slp->id) < 0) {
		log("unlink failed: %s", strerror(errno));
		return(-1);
	}
	if (slp == slist)
		slist = slp->next;
	else
		slp->prev->next = slp->next;
	if (slp->next != NULL)
		slp->next->prev = slp->prev;
	log("ID %d unlinked", slp->id);
	free(slp);
	return(0);
}

static int
ifdown(int id)
{
	pid_t pid;
	int status;
	char ifname[IFNAMSIZ];

	sprintf(ifname, "sl%d", id);
	if ((pid = fork()) < 0) {
		log("fork failed: %s", strerror(errno));
		return(-1);
	}
	if (pid == 0) {		/* child */
		execl(IFCONFIGPATH, IFCONFIG, ifname, "down", NULL);
		log("unexpected return from exec");
		exit(1);
	}
	/* parent */
	while (waitpid(pid, &status, 0) != pid)
		;
	if (WIFEXITED(status) && (WEXITSTATUS(status) == 0)) {
		return(0);
	} else {
		errno = EPROTO; /* as good as any */
		return(-1);
	}
}

static void
svc_driver(int drvfd)
{
	slink_t *slp;
	dev_t dev;
	int n;

	for (;;) {
		n = read(drvfd, &dev, sizeof(dev_t));
		if (n < 0) {
			if (errno == EAGAIN) {
				break;
			} else {
				log("read from driver failed: %s",
				  strerror(errno));
				exit(1);
			}
		}
		log("read dev 0x%x from driver", dev);
		for (slp = slist; slp != NULL; slp = slp->next)
			if (slp->dev == dev)
				break;
		if (slp == NULL)
			log("no slink_t found");
		else
			_discon(slp);
	}
}

static void
timeout(int signo)
{
	log("request timed out");
}
