/*
 * Copyright (c) 1993 Addison-Wesley Publishing Company.
 * All rights reserved.  This source code is provided
 * for its instructional value, and is not guaranteed
 * for any particular purpose.
 */
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

int
setup(char *device, char *initprog, char *args[])
{
	int fds[2];
	int i, nfd, narg, cstat;
	pid_t pid;
	struct iovec iov;
	struct msghdr msg;
	char **eargs;

	/*
	 * Create a pair of interconnected sockets.
	 */
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, fds) < 0)
		return(-1);

	/*
	 * Fork off a child process to execute the
	 * device-specific initialization program.
	 */
	if ((pid = fork()) < 0) {
		/*
		 * fork failed.  Close the pair of
		 * sockets just created.
		 */
		close(fds[0]);
		close(fds[1]);
		return(-1);
	} else if (pid == 0) {
		/*
		 * Child processing.  Close the parent's socket
		 * descriptor and file descriptor 0.  Then dup
		 * the child's socket to fd 0.  This gives the
		 * program a well-known file descriptor to use
		 * when sending the device file descriptor.  We
		 * don't need fds[1] after it is duplicated.
		 */
		close(fds[0]);
		close(0);
		if (dup(fds[1]) != 0)
			exit(1);
		close(fds[1]);

		/*
		 * Count the number of arguments and allocate
		 * enough space for the pointers plus the command
		 * name, the device name, and a terminating NULL
		 * pointer.
		 */
		for (narg = 0; args[narg] != NULL; narg++)
			;
		eargs = malloc((narg+3)*sizeof(char *));
		if (eargs == NULL)
			exit(1);

		/*
		 * The first argument is the basename of the
		 * command to be executed.
		 */
		eargs[0] = strrchr(initprog, '/');
		if (eargs[0] == NULL)
			eargs[0] = initprog;
		else
			eargs[0]++;

		/*
		 * Copy the other arguments.
		 */
		for (i = 0; i < narg; i++)
			eargs[i+1] = args[i];

		/*
		 * The last argument is the name of the device.
		 * End the array with a NULL pointer.
		 */
		eargs[++i] = device;
		eargs[++i] = NULL;

		/*
		 * Execute the specified program.
		 */
		execvp(initprog, eargs);
		exit(1);
	} else {
		/*
		 * The parent has no use for the child's socket
		 * descriptor, so close it.
		 */
		close(fds[1]);

		/*
		 * Wait for the child to terminate.
		 */
		if (waitpid(pid, &cstat, 0) < 0) {
			close(fds[0]);
			return(-1);
		}

		/*
		 * If the child terminated because of a signal,
		 * set errno to EINTR.  Otherwise, if the child
		 * terminated on its own with an unsuccessful
		 * status, set errno to EPROTO.
		 */
		if (WIFSIGNALED(cstat)) {
			close(fds[0]);
			errno = EINTR;
			return(-1);
		} else if (WIFEXITED(cstat)  &&
		  (WEXITSTATUS(cstat) != 0)) {
			close(fds[0]);
			errno = EPROTO;
			return(-1);
		}

		/*
		 * Prepare to receive the file descriptor.
		 */
		iov.iov_base = NULL;
		iov.iov_len = 0;
		msg.msg_iov = &iov;
		msg.msg_iovlen = 1;
		msg.msg_name = NULL;
		msg.msg_accrights = (caddr_t)&nfd;
		msg.msg_accrightslen = sizeof(nfd);

		/*
		 * If recvmsg fails, make sure the file
		 * descriptor is invalid.
		 */
		if (recvmsg(fds[0], &msg, 0) < 0)
			nfd = -1;

		/*
		 * Close the socket descriptor and return the
		 * file descriptor for the device.
		 */
		close(fds[0]);
		return(nfd);
	}
}
